/*
 *                            P R O T O . C
 *
 *  Data structures which control the protocol.
 *
 *  Version      : $Revision: 1.2 $
 *
 *  Created      : Fri May 20 21:35:00 1994
 *  Author       : Ulrich Drepper <drepper@mydec>
 *
 *  Last modified: Mon Jul  4 14:14:18 1994
 *  Author       : Ulrich Drepper <drepper@mydec>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 1, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
#if !defined(lint)
static const char *vcid = "$Id: proto.c,v 1.2 1994/07/04 13:14:24 drepper Exp $";
#endif /*lint */

#include "proto.h"

/*
 * exported variables
 */
struct EmpCommand empCommand[] = {
    { "user",    USER },
    { "coun",    COUN },
    { "quit",    QUIT },
    { "pass",    PASS },
    { "play",    PLAY },
    { "list",    LIST },
    { "cmd",     CMD  },
    { "ctld",    CTLD },
    { "wat",     WAT  },
    { "version", VERS }
};

/*
 * Local Variables:
 *  mode:c
 *  c-indent-level:4
 *  c-continued-statement-offset:4
 *  c-continued-brace-offset:0
 *  c-brace-offset:0
 *  c-imaginary-offset:0
 *  c-argdecl-indent:4
 *  c-label-offset:-2
 * End:
 */
